LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

LIBRARY dyplo;
USE dyplo.dyplo_components.ALL;

LIBRARY topic;
USE topic.topic_pkg.ALL;
USE topic.topic_tb_pkg.ALL;

USE std.env.ALL;

ENTITY dyplo_axis_data_packer_tb IS
END dyplo_axis_data_packer_tb;

ARCHITECTURE testbench OF dyplo_axis_data_packer_tb IS

	CONSTANT c_number_of_clocks : natural                                   := 1;
	CONSTANT c_clk_periods      : t_time_array(0 TO c_number_of_clocks - 1) := (10 ns, OTHERS => 0 ns);
	CONSTANT c_user_width       : natural                                   := 4;

	SIGNAL gsr         : std_logic;
	SIGNAL clocks      : std_logic_vector(c_number_of_clocks - 1 DOWNTO 0);
	SIGNAL resets      : std_logic_vector(c_number_of_clocks - 1 DOWNTO 0);
	SIGNAL resets_done : std_logic;
	SIGNAL stop_clocks : std_logic := '0';

	ALIAS clk : std_logic IS clocks(0);
	ALIAS rst : std_logic IS resets(0);

	SIGNAL data_in  : std_logic_vector(63 DOWNTO 0);
	SIGNAL strb_in  : std_logic_vector(1 DOWNTO 0) := (OTHERS => '1');
	SIGNAL user_in  : std_logic_vector(c_user_width - 1 DOWNTO 0);
	SIGNAL valid_in : std_logic;
	SIGNAL ready_in : std_logic;

	SIGNAL data_out  : std_logic_vector(63 DOWNTO 0);
	SIGNAL strb_out  : std_logic_vector(1 DOWNTO 0);
	SIGNAL user_out  : std_logic_vector(c_user_width - 1 DOWNTO 0);
	SIGNAL valid_out : std_logic;
	SIGNAL ready_out : std_logic;

	SIGNAL data_out_2  : std_logic_vector(63 DOWNTO 0);
	SIGNAL strb_out_2  : std_logic_vector(1 DOWNTO 0);
	SIGNAL user_out_2  : std_logic_vector(c_user_width - 1 DOWNTO 0);
	SIGNAL valid_out_2 : std_logic;
	SIGNAL ready_out_2 : std_logic;

BEGIN

	PROCESS
		VARIABLE v_transfer_cnt : natural := 0;
	BEGIN
		-- Wait for GSR / reset
		WAIT UNTIL resets_done = '1';
		WAIT UNTIL rising_edge(clk);

		WHILE v_transfer_cnt < 100 LOOP

			IF (valid_out_2 = '1' AND ready_out_2 = '1') THEN
				v_transfer_cnt := v_transfer_cnt + 1;
			END IF;

			WAIT UNTIL rising_edge(clk);

		END LOOP;

		stop_clocks <= '1';
		finish(0);

		WAIT;

	END PROCESS;

	check_input : PROCESS
		VARIABLE v_change_tuser_delay_cnt : natural range 0 to 10 := 0;
	BEGIN
		
		-- Change TUSER bits with delay
		IF(v_change_tuser_delay_cnt = 0) then
			user_in    <= get_random_slv(16, 4);
			v_change_tuser_delay_cnt := 10;
		ELSE
			v_change_tuser_delay_cnt := v_change_tuser_delay_cnt - 1;
		END IF;
		
		-- Change TSTRB bits
		strb_in(1) <= get_random_slv(4, 1)(0);

		WAIT UNTIL valid_in = '1' AND ready_in = '1';
		WAIT UNTIL rising_edge(clk);

	END PROCESS;

	traffic_gen_i : axis_traffic_gen
		GENERIC MAP(
			p_data_width             => 64,
			p_insert_id_in_data      => false,
			p_data_source            => "COUNTER",
			p_max_valid_window_width => 4,
			p_max_gap_width          => 4
		)
		PORT MAP(
			i_clk         => clk,
			i_rst         => rst,
			M_AXIS_TDATA  => data_in,
			M_AXIS_TREADY => ready_in,
			M_AXIS_TVALID => valid_in
		);

	dut_i : dyplo_axis_data_packer
		GENERIC MAP(
			p_tuser_width => c_user_width
		)
		PORT MAP(
			i_clk       => clk,
			i_rst       => rst,
			i_data_in   => data_in,
			i_strb_in   => strb_in,
			i_user_in   => user_in,
			i_valid_in  => valid_in,
			o_ready_in  => ready_in,
			o_data_out  => data_out,
			o_strb_out  => strb_out,
			o_user_out  => user_out,
			o_valid_out => valid_out,
			i_ready_out => ready_out
		);

	verify_i : axis_verifier
		GENERIC MAP(
			p_data_width             => 64,
			p_max_ready_window_width => 3,
			p_max_gap_width          => 5,
			p_check_data             => false
		)
		PORT MAP(
			i_clk             => clk,
			i_rst             => rst,
			S_AXIS_EXP_TDATA  => (OTHERS => '0'),
			S_AXIS_EXP_TREADY => OPEN,
			S_AXIS_EXP_TVALID => '1',
			S_AXIS_ACT_TDATA  => data_out,
			S_AXIS_ACT_TREADY => ready_out,
			S_AXIS_ACT_TVALID => valid_out
		);

	clk_rst_mng : clock_reset_manager
		GENERIC MAP(
			p_number_of_output_clocks => 1,
			p_clk_periods             => c_clk_periods,
			p_enable_gsr              => true,
			p_gsr_duration            => 100 ns
		)
		PORT MAP(
			o_gsr         => gsr,
			o_clks        => clocks,
			o_rsts        => resets,
			o_rsts_n      => OPEN,
			o_resets_done => resets_done,
			i_stop_clocks => stop_clocks
		);

END testbench;
