-- File: tb_dyplo_cdc_fifo_af.vhd
--
-- (c) Copyright 2012 - 2013 Topic Embedded Systems. All rights reserved.
--
-- This file contains confidential and proprietary information of Topic
-- Embedded Systems and is protected under U.S. and international copyright
-- and other intellectual property laws.
--
-- DISCLAIMER
-- This disclaimer is not a license and does not grant any rights to the
-- materials distributed herewith. Except as otherwise provided in a valid
-- license issued to you by Topic Embedded Systems, and to the maximum extent
-- permitted by applicable law:
-- (1) THESE MATERIALS ARE MADE AVAILABLE "AS IS" AND WITH ALL FAULTS,
--     AND TOPIC EMBEDDED SYSTEMS HEREBY DISCLAIMS ALL WARRANTIES AND
--     CONDITIONS, EXPRESS, IMPLIED, OR STATUTORY, INCLUDING BUT NOT LIMITED
--     TO WARRANTIES OF MERCHANTABILITY, NON-INFRINGEMENT, OR FITNESS FOR ANY
--     PARTICULAR PURPOSE;
-- and
-- (2) Topic Embedded Systems shall not be liable (whether in contract or tort,
--     including negligence, or under any other theory of liability) for any
--     loss or damage of any kind or nature related to, arising under or in
--     connection with these materials, including for any direct, or any
--     indirect, special, incidental, or consequential loss or damage
--     (including loss of data, profits, goodwill, or any type of loss or
--     damage suffered as a result of any action brought by a third party)
--     even if such damage or loss was reasonably foreseeable or Topic Embedded
--     Systems had been advised of the possibility of the same.
--
-- CRITICAL APPLICATIONS
-- Topic Embedded System products are not designed or intended to be fail-safe,
-- or for use in any application requiring fail-safe performance, such as
-- life-support or safety devices or systems, Class III medical devices,
-- nuclear facilities, applications related to the deployment of airbags, or
-- any other applications that could lead to death, personal injury, or severe
-- property or environmental damage (individually and collectively, "Critical
-- Applications"). Customer assumes the sole risk and liability of any use of
-- Topic Embedded Systems products in Critical Applications, subject only to
-- applicable laws and regulations governing limitations on product liability.
--
-- THIS COPYRIGHT NOTICE AND DISCLAIMER MUST BE RETAINED AS PART OF THIS FILE
-- AT ALL TIMES.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use std.env.all;

library dyplo;
use dyplo.dyplo_types.all;
use dyplo.dyplo_config.all;
use dyplo.dyplo_components.all;

entity tb_dyplo_cdc_fifo_af is
   generic(
      DATA_WIDTH : integer := 32;
      FIFO_TYPE  : string  := "DISTRIBUTED"
   );
end entity;

architecture testbench of tb_dyplo_cdc_fifo_af is
   constant c_clk_period_wr_clk : time := 6.667 ns;
   constant c_clk_period_rd_clk : time := 11.0 ns;

   signal wr_clk_i   : std_logic := '0';
   signal rd_clk_i   : std_logic := '0';
   signal GSR        : std_logic := '1';
   signal wr_reset_i : std_logic := '1';
   signal rd_reset_i : std_logic := '1';

   signal wr_data_i     : std_logic_vector(DATA_WIDTH - 1 downto 0);
   signal wr_en_i       : std_logic := '0';
   signal wr_en         : std_logic := '0';
   signal full_i        : std_logic;
   signal almost_full_i : std_logic;
   signal rd_data_i     : std_logic_vector(DATA_WIDTH - 1 downto 0);
   signal rd_en_i       : std_logic := '0';
   signal rd_en         : std_logic := '0';
   signal empty_i       : std_logic;

   signal data_in_counter : integer := 0;

   signal end_simulation : std_logic := '0';

   shared variable data_send : integer := 0;

   signal write_data_done : std_logic := '0';
   signal read_data_done  : std_logic := '0';

   signal test_step : natural := 0;

begin

   -- Global Set Rest (GSR)
   process
   begin
      wait for 100 ns;
      GSR <= '0';
      wait;
   end process;

   -- Create clock after Global Set Reset is finished
   wr_clk_i <= not wr_clk_i after c_clk_period_wr_clk / 2 when GSR = '0';
   rd_clk_i <= not rd_clk_i after c_clk_period_rd_clk / 2 when GSR = '0';

   -- Overall process to start and stop testcases
   process
   begin
      end_simulation <= '0';
      test_step      <= 1;

      -- wait until GSR is deasserted
      wait until GSR = '0';

      -- for loop to wait 6 clock cycles, then de-activate reset
      for i in 0 to 6 loop
         wait until rising_edge(rd_clk_i);
      end loop;
      rd_reset_i <= '0';

      -- for loop to wait 6 clock cycles, then de-activate reset
      for i in 0 to 6 loop
         wait until rising_edge(wr_clk_i);
      end loop;
      wr_reset_i <= '0';

      wait until rising_edge(wr_clk_i);

      -- Fill fifo with 32 words
      wait until write_data_done = '1';
      if (almost_full_i /= '1') then
         report "Almost_full flag should be asserted by now, but isn't" severity failure;
      end if;
      test_step <= test_step + 1;       -- 2

      -- Empty fifo 
      wait until read_data_done = '1';

      -- for loop to wait for the delay in almost_full deassertion
      for i in 0 to 6 loop
         wait until rising_edge(rd_clk_i); -- wait for 6 rd_clk ticks because that is the slowest clock
         if (almost_full_i = '0') then
            exit;
         end if;
      end loop;

      if (almost_full_i /= '0') then
         report "Almost_full flag should be de-asserted by now, but isn't" severity failure;
      end if;

      test_step <= test_step + 1;       -- 3
      wait until rising_edge(wr_clk_i);
      wait until rising_edge(rd_clk_i);

      -- Fill fifo completely
      wait until write_data_done = '1';
      if (almost_full_i /= '1' or full_i /= '1') then
         report "Almost_full and full flag should be asserted by now, but at least 1 of them isn't" severity failure;
      end if;
      test_step <= test_step + 1;       -- 4

      -- Empty fifo 
      wait until read_data_done = '1';
      if (empty_i /= '1' or almost_full_i /= '0') then
         report "The empty flag or almost_full flag do not have the expected state" severity failure;
      end if;

      end_simulation <= '1';

      -- Make sure that simulation finishes  
      wait until rising_edge(wr_clk_i);
      if (end_simulation = '1') then
         report "End of simulation";
         finish(0);
      end if;

      wait;
   end process;

   data_in_interface : process(wr_clk_i)
      variable full_assertion_time : natural := 0;
   begin
      if (rising_edge(wr_clk_i)) then
         if (wr_reset_i = '1') then
            wr_en_i <= '0';
         else
            case (test_step) is
               -- fill fifo with 32 words
               when 1 =>
                  if (full_i = '0') then
                     wr_en_i <= '1';
                  else
                     wr_en_i <= '0';
                  end if;

                  if (wr_en_i = '1' and full_i = '0') then
                     data_in_counter <= data_in_counter + 1;
                  end if;

                  if (data_in_counter = 64) then
                     wr_en_i         <= '0';
                     write_data_done <= '1';
                  end if;

               -- fill fifo completely
               when 3 =>
                  write_data_done <= '0';
                  if (full_i = '0') then
                     wr_en_i <= '1';
                  else
                     wr_en_i <= '0';
                  end if;

                  if (wr_en_i = '1' and full_i = '0') then
                     data_in_counter <= data_in_counter + 1;
                  end if;

                  if (data_in_counter > (2 * 64) + 2) then
                     if (almost_full_i = '0') then
                        report "The almost_full flag needs to be asserted now" severity failure;
                     end if;
                  end if;

                  if (full_i = '1') then
                     full_assertion_time := full_assertion_time + 1;
                  else
                     full_assertion_time := 0;
                  end if;

                  if (full_assertion_time >= 16) then
                     wr_en_i         <= '0';
                     write_data_done <= '1';
                  end if;

               when others =>
                  null;
            end case;

            if (wr_en_i = '1' and full_i = '0') then
               data_in_counter <= data_in_counter + 1;
            end if;

         end if;
      end if;
   end process data_in_interface;

   wr_data_i <= std_logic_vector(to_unsigned(data_in_counter, DATA_WIDTH));
   wr_en     <= wr_en_i and (not full_i);

   data_out_interface : process(rd_clk_i)
      variable data_out_counter : integer := 0;
   begin
      if (rising_edge(rd_clk_i)) then
         if (rd_reset_i = '1') then
            rd_en_i <= '0';
         else
            case (test_step) is
               -- empty fifo
               when 2 =>
                  if (empty_i = '0') then
                     rd_en_i <= '1';
                  else
                     rd_en_i <= '0';
                  end if;

                  if (data_out_counter = 64) then
                     read_data_done <= '1';
                  end if;

               -- empty fifo
               when 4 =>
                  read_data_done <= '0';
                  if (empty_i = '0') then
                     rd_en_i <= '1';
                  else
                     rd_en_i <= '0';
                  end if;

                  if (data_out_counter < 64) then
                     if (almost_full_i = '0') then
                        report "The almost_full flag needs to be asserted now" severity failure;
                     end if;
                  end if;

                  if (data_out_counter = data_in_counter) then
                     read_data_done <= '1';
                  end if;

               when others =>
                  null;
            end case;

            if (rd_en_i = '1' and empty_i = '0') then
               if (rd_data_i /= std_logic_vector(to_unsigned(data_out_counter, DATA_WIDTH))) then
                  report "Data does not match expected data" severity failure;
               end if;
               data_out_counter := data_out_counter + 1;
            end if;

         end if;
      end if;
   end process data_out_interface;

   rd_en <= rd_en_i and (not empty_i);

   dut : dyplo_cdc_fifo_af
      generic map(
         DATA_WIDTH => DATA_WIDTH,
         ADDR_WIDTH => 7,
         FIFO_TYPE  => FIFO_TYPE
      )
      port map(
         wr_clk      => wr_clk_i,
         wr_reset    => wr_reset_i,
         wr_data     => wr_data_i,
         wr_en       => wr_en,
         full        => full_i,
         almost_full => almost_full_i,
         rd_clk      => rd_clk_i,
         rd_reset    => rd_reset_i,
         rd_data     => rd_data_i,
         rd_en       => rd_en,
         empty       => empty_i
      );

end testbench;
