-- File: tb_dyplo_connect_slice.vhd
--
-- (c) Copyright 2012 - 2013 Topic Embedded Systems. All rights reserved.
--
-- This file contains confidential and proprietary information of Topic
-- Embedded Systems and is protected under U.S. and international copyright
-- and other intellectual property laws.
--
-- DISCLAIMER
-- This disclaimer is not a license and does not grant any rights to the
-- materials distributed herewith. Except as otherwise provided in a valid
-- license issued to you by Topic Embedded Systems, and to the maximum extent
-- permitted by applicable law:
-- (1) THESE MATERIALS ARE MADE AVAILABLE "AS IS" AND WITH ALL FAULTS,
--     AND TOPIC EMBEDDED SYSTEMS HEREBY DISCLAIMS ALL WARRANTIES AND
--     CONDITIONS, EXPRESS, IMPLIED, OR STATUTORY, INCLUDING BUT NOT LIMITED
--     TO WARRANTIES OF MERCHANTABILITY, NON-INFRINGEMENT, OR FITNESS FOR ANY
--     PARTICULAR PURPOSE;
-- and
-- (2) Topic Embedded Systems shall not be liable (whether in contract or tort,
--     including negligence, or under any other theory of liability) for any
--     loss or damage of any kind or nature related to, arising under or in
--     connection with these materials, including for any direct, or any
--     indirect, special, incidental, or consequential loss or damage
--     (including loss of data, profits, goodwill, or any type of loss or
--     damage suffered as a result of any action brought by a third party)
--     even if such damage or loss was reasonably foreseeable or Topic Embedded
--     Systems had been advised of the possibility of the same.
--
-- CRITICAL APPLICATIONS
-- Topic Embedded System products are not designed or intended to be fail-safe,
-- or for use in any application requiring fail-safe performance, such as
-- life-support or safety devices or systems, Class III medical devices,
-- nuclear facilities, applications related to the deployment of airbags, or
-- any other applications that could lead to death, personal injury, or severe
-- property or environmental damage (individually and collectively, "Critical
-- Applications"). Customer assumes the sole risk and liability of any use of
-- Topic Embedded Systems products in Critical Applications, subject only to
-- applicable laws and regulations governing limitations on product liability.
--
-- THIS COPYRIGHT NOTICE AND DISCLAIMER MUST BE RETAINED AS PART OF THIS FILE
-- AT ALL TIMES.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library std;
use std.env.all;

library dyplo;
use dyplo.dyplo_types.all;
use dyplo.dyplo_config.all;
use dyplo.dyplo_components.all;

library dyplo_test;
use dyplo_test.test_types.all;
use dyplo_test.test_components.all;

entity tb_dyplo_connect_slice is
	generic(
		CONNECT_ID    : integer := 2;
		CONNECT_LANES : integer := 4;
		NODES         : integer := 8;
		MONITOR       : boolean := true
	);
end tb_dyplo_connect_slice;

architecture testbench of tb_dyplo_connect_slice is
	
	constant c_clk_period_bp_clk : time := 6.667 ns; -- 150 MHz backplane clock
	constant c_stop_test_pulse_length  : integer := 6;

	signal bp_clk_i : std_logic := '0';
	signal GSR      : std_logic := '1';

	signal bp_ena_i : std_logic := '0';

	signal stream_destinations_i : destination_matrix(0 to 3) := (others => (others => '0'));

	signal f2b_tdata_i      : std_logic_vector(63 downto 0) := (others => '0');
	signal f2b_tstrb_i 		: std_logic_vector(1 downto 0)  := (others => '1');
	signal f2b_tuser_i      : std_logic_vector(3 downto 0)  := (others => '0');
	signal f2b_tstream_id_i : std_logic_vector(1 downto 0)  := (others => '0');
	signal f2b_tvalid_i     : std_logic                     := '0';
	signal f2b_tready_i     : std_logic                     := '0';

	signal b2f_tdata_i      : std_logic_vector(63 downto 0) := (others => '0');
	signal b2f_tstrb 		   : std_logic_vector(1 downto 0);
	signal b2f_tuser_i      : std_logic_vector(3 downto 0)  := (others => '0');
	signal b2f_tstream_id_i : std_logic_vector(1 downto 0)  := (others => '0');
	signal b2f_tvalid_i     : std_logic                     := '0';

	signal node_status_flags_i : std_logic_vector(c_status_flags_width - 1 downto 0) := (others => '0');
	signal dest_fifo_status_i  : std_logic_vector(c_status_flags_width - 1 downto 0) := (others => '0');

	signal status_lanes_in_i  : std_logic_vector(c_status_lanes_width - 1 downto 0) := (others => '0');
	signal status_lanes_out_i : std_logic_vector(c_status_lanes_width - 1 downto 0) := (others => '0');

	signal data_lanes_in_i  : data_lanes_type(0 to CONNECT_LANES - 1);
	signal data_lanes_out_i : data_lanes_type(0 to CONNECT_LANES - 1);

	signal lanes_mode_i        : lane_mode_type(0 to CONNECT_LANES - 1)         := (others => 0);
	signal lanes_destination_i : lane_destinations_type(0 to CONNECT_LANES - 1) := (others => (others => '0'));

	signal status_error_i : std_logic := '0';

	signal f2b_count_i : std_logic_vector(31 downto 0) := (others => '0');
	signal b2f_count_i : std_logic_vector(31 downto 0) := (others => '0');
	signal bp_count_i  : std_logic_vector(31 downto 0) := (others => '0');

	signal testcase    : natural   := 0;
	signal reset_test  : std_logic := '0';
	signal stop_inputs : std_logic := '0';

	signal send_to_lanes_i     : unsigned_matrix_4_8_4_64 := (others => (others => (others => (others => '0'))));
	signal captured_on_lanes_i : unsigned_matrix_8_4_64   := (others => (others => (others => '0')));

	signal send_to_f2b     : unsigned_matrix_4_64 := (others => (others => '0'));
	signal captured_on_b2f : unsigned_matrix_4_64 := (others => (others => '0'));

begin

	-- Global Set Rest (GSR)
	process
	begin
		wait for 100 ns;
		GSR <= '0';
		wait;
	end process;

	-- Create clock after Global Set Reset is finished
	bp_clk_i <= not bp_clk_i after c_clk_period_bp_clk / 2 when GSR = '0';

	-- Overall process to start and stop testcases
	process
	begin
		-- wait until GSR is deasserted
		wait until GSR = '0';

		-- for loop to wait 6 clock cycles, then deassert reset
		for i in 0 to 6 loop
			wait until rising_edge(bp_clk_i);
		end loop;

		-- Reset test 
		reset_test  <= '1';
		stop_inputs <= '0';
		wait until rising_edge(bp_clk_i);
		reset_test <= '0';

		--------------------------------------------------------
		-- TestCase 1:
		-- Test if no data is being accepted and no data is coming out of the connect slice if
		-- the bp_enable signal is deasserted.		
		report "Testcase 1 started";
		report "This test tests the influence of the bp_ena signal";
		testcase <= 1;
		for i in 0 to 50 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		stop_inputs <= '1';
		for i in 0 to c_stop_test_pulse_length loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';

		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
		stop_inputs <= '0';
      report "Testcase 1 finished";

		----------------------------------------------------------
		-- TestCase 2:	
		-- Send data from 1 lane to the b2f interface 		
		report "Testcase 2 started";
      report "Send data from 1 lane to the b2f interface";
		testcase <= 2;
		for i in 0 to 100 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		stop_inputs <= '1';
		for i in 0 to c_stop_test_pulse_length loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';
		
		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
		stop_inputs <= '0';
      report "Testcase 2 finished";	
		
		-----------------------------------------------------------
		-- TestCase 3:	
		-- Check status lanes out 		
      report "Testcase 3 started";
      report "Check status lanes output"; 
		testcase <= 3;
		for i in 0 to 200 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';
		
		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
      report "Testcase 3 finished";   
		
		---------------------------------------------------------
		-- TestCase 4:	
		-- Check dest_fifo_status signals 		
      report "Testcase 4 started";
      report "Check dest_fifo_status signals";
		testcase <= 4;
		for i in 0 to 120 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';
		
		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
      report "Testcase 4 finished";   
		
		------------------------------------------------------
		-- TestCase 5:	
      -- Send data to f2b interface and check on output lanes
      report "Testcase 5 started";
      report "Send data to f2b interface and check output lanes";
      testcase <= 5;
      for i in 0 to 1000 loop
         wait until rising_edge(bp_clk_i);
      end loop;
		stop_inputs <= '1';
		for i in 0 to c_stop_test_pulse_length loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';
		
		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
		stop_inputs <= '0';
      report "Testcase 5 finished";  
      
      ------------------------------------------------------
		-- TestCase 6:	
      -- Send data to f2b interface while pushing data on the lanes with the same destination node_id
      report "Testcase 6 started";
      report "Send data to f2b interface and on the lanes with the same destination node_id";
      testcase <= 6;
      for i in 0 to 600 loop
         wait until rising_edge(bp_clk_i);
      end loop;
		stop_inputs <= '1';
		for i in 0 to c_stop_test_pulse_length loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';
		
		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
		stop_inputs <= '0';
      report "Testcase 6 finished";
      
      ------------------------------------------------------
		-- TestCase 7:	
      -- Send data to f2b interface and read from b2f while pushing data on the lanes other lanes
      report "Testcase 7 started";
      report "Send data to f2b interface and read from b2f interface while pushing data";
      testcase <= 7;
      for i in 0 to 600 loop
         wait until rising_edge(bp_clk_i);
      end loop;
		stop_inputs <= '1';
		for i in 0 to c_stop_test_pulse_length loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test <= '1';
		
		-- Reset test (last some cycles to clear all pipelines)
		for i in 0 to 3 loop
			wait until rising_edge(bp_clk_i);
		end loop;
		reset_test 	<= '0';
		stop_inputs <= '0';
      report "Testcase 7 finished";   
      
      ------------------------------------------------------
      -- TestCase 8:   
      -- Send data to f2b interface and read from b2f while pushing a lot of data on the lanes
      report "Testcase 8 started";
      report "Send data to f2b interface and read from b2f interface while pushing a lot of data";
      testcase <= 8;
      for i in 0 to 2000 loop
         wait until rising_edge(bp_clk_i);
      end loop;
      stop_inputs <= '1';
      for i in 0 to c_stop_test_pulse_length loop
         wait until rising_edge(bp_clk_i);
      end loop;
      reset_test <= '1';
      
      -- Reset test (last some cycles to clear all pipelines)
      for i in 0 to 3 loop
         wait until rising_edge(bp_clk_i);
      end loop;
      reset_test  <= '0';
      stop_inputs <= '0';
      report "Testcase 8 finished";  
      
      ------------------------------------------------------
      -- TestCase 9:   
      -- Send data to f2b interface and read from b2f while fully occupying all lanes
      report "Testcase 9 started";
      report "Send data to f2b interface and read from b2f interface while pushing a lot of data";
      testcase <= 9;
      for i in 0 to 2000 loop
         wait until rising_edge(bp_clk_i);
      end loop;
      stop_inputs <= '1';
      for i in 0 to c_stop_test_pulse_length loop
         wait until rising_edge(bp_clk_i);
      end loop;
      reset_test <= '1';
      
      -- Reset test (last some cycles to clear all pipelines)
      for i in 0 to 3 loop
         wait until rising_edge(bp_clk_i);
      end loop;
      reset_test  <= '0';
      stop_inputs <= '0';
      report "Testcase 9 finished";            
      
      
      -- Make sure that simulation finishes  
      testcase <= 10;  
      wait until rising_edge(bp_clk_i); 
      if(testcase = 10) then
      	report "End of simulation";
      	finish(0);
      end if;
      
		wait;
	end process;
	
	bp_ena_interface : process(bp_clk_i)
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				bp_ena_i <= '0';
			else
				case (testcase) is
					when 1 =>
						bp_ena_i <= '0';
					when 2 | 5 | 6 | 7 | 8 | 9 =>
						bp_ena_i <= '1';					
					when others =>
						null;
				end case;
			end if;
		end if;
	end process bp_ena_interface;	
	
	stream_destinations_interface : process(bp_clk_i)
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				stream_destinations_i <= (others => (others => '0')); 
			else
				case (testcase) is
					---------------------------------------------
					when 4 =>						--  node_id & stream_id
						stream_destinations_i(0) <= "00000" & "11"; 
						stream_destinations_i(1) <= "00001" & "10"; 
						stream_destinations_i(2) <= "00010" & "01"; 
						stream_destinations_i(3) <= "00100" & "00";
					---------------------------------------------
					when 5 =>
						stream_destinations_i(0) <= "00001" & "10";
					----------------------------------------------	
					when 6 =>
						stream_destinations_i(0) <= "00001" & "10";
						stream_destinations_i(1) <= "00011" & "11";
					----------------------------------------------	
					when 7 =>
						stream_destinations_i(0) <= "00001" & "11";
						stream_destinations_i(1) <= "00111" & "10";   
						stream_destinations_i(2) <= "00011" & "01"; 
						stream_destinations_i(3) <= "00101" & "00";
					----------------------------------------------	
					when 8 =>
						stream_destinations_i(0) <= "00011" & "11";
						stream_destinations_i(1) <= "00110" & "11";   
						stream_destinations_i(2) <= "00000" & "01"; 
						stream_destinations_i(3) <= "00101" & "01";
					----------------------------------------------	
					when 9 =>  
						stream_destinations_i(2) <= "00010" & "01";  
					when others =>
						null;
				end case;

			end if;
		end if;
	end process stream_destinations_interface;	

	f2b_interface : process(bp_clk_i, send_to_f2b, f2b_tstream_id_i)
		variable cnt : integer := 0;
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				f2b_tvalid_i     	<= '0';
				f2b_tuser_i      	<= (others => '0');
				f2b_tstream_id_i 	<= "00";
				send_to_f2b   		<= (others => (others => '0'));
				cnt              	:= 0;
			else
				case (testcase) is
               ------------------------------
					when 1 =>
						f2b_tvalid_i <= '1';
					----------------------------	
               when 5 =>
                  if(cnt = 5) then -- small delay, then start outputting data on stream 0
                  	f2b_tvalid_i 		<= '1';
                  	f2b_tstream_id_i 	<= "00";
                  end if;
                  
                  if(stop_inputs = '1') then
                     f2b_tvalid_i <= '0';
                  end if;
               -----------------------------	
               when 6 =>
                  if(cnt = 5) then -- small delay, then start outputting data on stream 0 and 1
                  	f2b_tvalid_i 		<= '1';
                  end if;
                  
                  if(stop_inputs = '1') then
                     f2b_tvalid_i <= '0';
                  end if;
                  
                  -- Continue switching between 00 and 01
                  if(f2b_tstream_id_i = "00") then
                  	f2b_tstream_id_i <= "01";
                  elsif(f2b_tstream_id_i = "01") then
                  	f2b_tstream_id_i 	<= "00";
                  end if;
                  
                  if(send_to_f2b(0) > 2) then
                  	report "No data should be sent on stream 0 (except for some pipeline filling) because input lane 1 is fully occupied with data for the same destination node (1)" severity failure;
                  end if;
               -----------------------------	
               when 7 =>
                  if(cnt = 5) then -- small delay, then start outputting data on all streams
                  	f2b_tvalid_i 		<= '1';
                  end if;
                  
                  if(stop_inputs = '1') then
                     f2b_tvalid_i <= '0';
                  end if;
                  
                  -- Continuous loop through streams
                  f2b_tstream_id_i <= std_logic_vector(unsigned(f2b_tstream_id_i) + 1);
               -----------------------------	
               when 8 =>
                  if(cnt = 5) then -- small delay, then start outputting data on all streams
                  	f2b_tvalid_i 		<= '1';
                  end if;
                  
                  if(stop_inputs = '1') then
                     f2b_tvalid_i <= '0';
                  end if;
                  
                  -- Continuous loop through streams
                  f2b_tstream_id_i <= std_logic_vector(unsigned(f2b_tstream_id_i) + 1);
               -----------------------------	
               when 9 =>
                  if(cnt = 5) then -- small delay, then start outputting data on all streams
                  	f2b_tvalid_i 		<= '1';
                  end if;
                  
                  if(stop_inputs = '1') then
                     f2b_tvalid_i <= '0';
                  end if;
                  
                  -- Only output stream 10 with destination node 2
                  f2b_tstream_id_i <= "10";
                  
               -----------------------------      
					when others =>
						null;
				end case;

				-- update input_cnt on transfer
				if (f2b_tvalid_i = '1' and f2b_tready_i = '1') then
					send_to_f2b(to_integer(unsigned(f2b_tstream_id_i))) <= send_to_f2b(to_integer(unsigned(f2b_tstream_id_i))) + 1;
				end if;
				
				cnt := cnt + 1;

			end if;
		end if;

		f2b_tdata_i <= std_logic_vector(send_to_f2b(to_integer(unsigned(f2b_tstream_id_i))));

	end process f2b_interface;

	b2f_interface : process(bp_clk_i)
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				captured_on_b2f 		<= (others => (others => '0'));
			else
				case (testcase) is
					when 1 =>
						if (b2f_tvalid_i = '1') then
							report "Signal 'b2f_tvalid' asserted while bp_ena is '0'" severity failure;
						end if;
						
					when 2 =>
						if (b2f_tvalid_i = '1' and b2f_tstream_id_i /= "10") then
							report "Signal 'b2f_tstream_id' is not 2 as expected" severity failure;
						end if;	
						
					when others =>
						null;
				end case;

				-- check and update output_cnt on transfer
				if (b2f_tvalid_i = '1') then
					if (std_logic_vector(captured_on_b2f(to_integer(unsigned(b2f_tstream_id_i)))) /= b2f_tdata_i) then
						report "Data on b2f interface does not match expected data" severity failure;
					end if;
					captured_on_b2f(to_integer(unsigned(b2f_tstream_id_i))) <= captured_on_b2f(to_integer(unsigned(b2f_tstream_id_i))) + 1;
				end if;

			end if;
		end if;
	end process b2f_interface;

	lanes_in_interface : process(bp_clk_i)
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				lanes_mode_i        <= (others => 0);
				lanes_destination_i <= (others => (others => '0'));
			else
				case (testcase) is
					----------------------------------------------------
					when 1 =>
						if(stop_inputs = '0') then	
							-- lane 0: fully occupy this lane, stream to node 1, stream 2
							lanes_mode_i(0)        <= 3;
							lanes_destination_i(0) <= std_logic_vector(to_unsigned(1, c_node_id_width)) & std_logic_vector(to_unsigned(2, c_stream_id_width));
	
							-- lane 1: occupy 1/10th of this lane, stream to node 4, stream 0
							lanes_mode_i(1)        <= 1;
							lanes_destination_i(1) <= std_logic_vector(to_unsigned(4, c_node_id_width)) & std_logic_vector(to_unsigned(0, c_stream_id_width));
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;
					---------------------------------------------------
					when 2 =>
						if(stop_inputs = '0') then	
							-- lane 0: fully occupy this lane, stream to node 2, stream 2
							lanes_mode_i(0)        <= 3;
							lanes_destination_i(0) <= std_logic_vector(to_unsigned(2, c_node_id_width)) & std_logic_vector(to_unsigned(2, c_stream_id_width));
	
							-- lane 1: fully occupy this lane, stream to node 4, stream 0
							lanes_mode_i(1)        <= 3;
							lanes_destination_i(1) <= std_logic_vector(to_unsigned(4, c_node_id_width)) & std_logic_vector(to_unsigned(0, c_stream_id_width));
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;	
					---------------------------------------------------
					when 5 =>
						if(stop_inputs = '0') then	
							-- lane 0: occupy 1/3th of this lane, stream to node 6, stream 0
							lanes_mode_i(0)        <= 2;
							lanes_destination_i(0) <= std_logic_vector(to_unsigned(6, c_node_id_width)) & std_logic_vector(to_unsigned(0, c_stream_id_width));
	
							-- lane 1: fully occupy this lane, stream to node 7, stream 1
							lanes_mode_i(1)        <= 3;
							lanes_destination_i(1) <= std_logic_vector(to_unsigned(7, c_node_id_width)) & std_logic_vector(to_unsigned(1, c_stream_id_width));
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;	
					---------------------------------------------------
					when 6 =>
						if(stop_inputs = '0') then	
                     -- lane 0: fully occupy this lane, stream to node 1, stream 0
                     lanes_mode_i(0)        <= 3;
                     lanes_destination_i(0) <= std_logic_vector(to_unsigned(1, c_node_id_width)) & std_logic_vector(to_unsigned(0, c_stream_id_width));
   
                     -- lane 1: occupy 1/3th of this lane, stream to node 3, stream 1
                     lanes_mode_i(1)        <= 2;
                     lanes_destination_i(1) <= std_logic_vector(to_unsigned(3, c_node_id_width)) & std_logic_vector(to_unsigned(1, c_stream_id_width));
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;	
					---------------------------------------------------
					when 7 =>
						if(stop_inputs = '0') then	
                     -- lane 0: fully occupy this lane, stream to node 0, stream 1
                     lanes_mode_i(0)        <= 3;
                     lanes_destination_i(0) <= std_logic_vector(to_unsigned(0, c_node_id_width)) & std_logic_vector(to_unsigned(1, c_stream_id_width));
   
                     -- lane 1: occupy 1/10th of this lane, stream to node 2, stream 1
                     lanes_mode_i(1)        <= 1;
                     lanes_destination_i(1) <= std_logic_vector(to_unsigned(2, c_node_id_width)) & std_logic_vector(to_unsigned(1, c_stream_id_width));
                     
                     -- lane 2: fully occupy this lane, stream to node 6, stream 2
                     lanes_mode_i(2)        <= 3;
                     lanes_destination_i(2) <= std_logic_vector(to_unsigned(6, c_node_id_width)) & std_logic_vector(to_unsigned(2, c_stream_id_width));
   
                     -- lane 3: occupy 1/3th of this lane, stream to node 4, stream 3
                     lanes_mode_i(3)        <= 2;
                     lanes_destination_i(3) <= std_logic_vector(to_unsigned(4, c_node_id_width)) & std_logic_vector(to_unsigned(3, c_stream_id_width));                     
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;	
					---------------------------------------------------
					when 8 =>
						if(stop_inputs = '0') then	
                     -- lane 0: fully occupy this lane, stream to node 4, stream 3
                     lanes_mode_i(0)        <= 3;
                     lanes_destination_i(0) <= std_logic_vector(to_unsigned(4, c_node_id_width)) & std_logic_vector(to_unsigned(3, c_stream_id_width));
   
                     -- lane 1: occupy 1/3th of this lane, stream to node 1, stream 1
                     lanes_mode_i(1)        <= 2;
                     lanes_destination_i(1) <= std_logic_vector(to_unsigned(1, c_node_id_width)) & std_logic_vector(to_unsigned(1, c_stream_id_width));
                     
                     -- lane 2: fully occupy this lane, stream to node 2, stream 2
                     lanes_mode_i(2)        <= 3;
                     lanes_destination_i(2) <= std_logic_vector(to_unsigned(2, c_node_id_width)) & std_logic_vector(to_unsigned(2, c_stream_id_width));
   
                     -- lane 3: occupy 1/3th of this lane, stream to node 7, stream 3
                     lanes_mode_i(3)        <= 2;
                     lanes_destination_i(3) <= std_logic_vector(to_unsigned(7, c_node_id_width)) & std_logic_vector(to_unsigned(3, c_stream_id_width));                     
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;	
					---------------------------------------------------
					when 9 =>
						if(stop_inputs = '0') then	
                     -- lane 0: fully occupy this lane, stream to node 4, stream 3
                     lanes_mode_i(0)        <= 3;
                     lanes_destination_i(0) <= std_logic_vector(to_unsigned(4, c_node_id_width)) & std_logic_vector(to_unsigned(3, c_stream_id_width));
   
                     -- lane 1: fully occupy this lane, stream to node 1, stream 1
                     lanes_mode_i(1)        <= 3;
                     lanes_destination_i(1) <= std_logic_vector(to_unsigned(1, c_node_id_width)) & std_logic_vector(to_unsigned(1, c_stream_id_width));
                     
                     -- lane 2: fully occupy this lane, stream to node 2, stream 2
                     lanes_mode_i(2)        <= 3;
                     lanes_destination_i(2) <= std_logic_vector(to_unsigned(2, c_node_id_width)) & std_logic_vector(to_unsigned(2, c_stream_id_width));
   
                     -- lane 3: fully occupy this lane, stream to node 7, stream 3
                     lanes_mode_i(3)        <= 3;
                     lanes_destination_i(3) <= std_logic_vector(to_unsigned(7, c_node_id_width)) & std_logic_vector(to_unsigned(3, c_stream_id_width));                     
						else
							-- Stop all input lanes
							lanes_mode_i	<= (others => 0);
						end if;					

					when others =>
						null;
				end case;
				
	         -- check if there are multiple data words on the in_lanes with the same destination node
            for i in 0 to CONNECT_LANES - 1 loop
               if (data_lanes_in_i(i).tvalid = '1') then
                  for j in 0 to CONNECT_LANES - 1 loop
                     if (j /= i) then
                        if (data_lanes_in_i(j).tvalid = '1' and (data_lanes_in_i(j).tnode_id = data_lanes_in_i(i).tnode_id)) then
                           report "There are multiple data words on the backplane lanes (input) (next to each other) with the same destination" severity failure;
                        end if;
                     end if;
                  end loop;
               end if;
            end loop;
		
			end if;
		end if;
	end process lanes_in_interface;

	lanes_out_interface : process(bp_clk_i)
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
			else
				case (testcase) is
					when others =>
						null;
				end case;

			end if;
		end if;
	end process lanes_out_interface;
	
	node_status_flags_interface : process(bp_clk_i)
		variable cnt : integer := 0;
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				node_status_flags_i <= (others => '0');
				cnt                 := 0;
			else
				case (testcase) is
               ------------------------------------
               when 3      =>
						if(cnt = 10) then
							node_status_flags_i <= "0001";
						elsif(cnt = 40) then
							node_status_flags_i <= "0101";
						elsif(cnt = 67) then
							node_status_flags_i <= "0100";
						elsif(cnt = 86) then
							node_status_flags_i <= "0000";
						elsif(cnt = 105) then
							node_status_flags_i <= "1010";
						elsif(cnt = 123) then
							node_status_flags_i <= "0011";
						elsif(cnt = 160) then
							node_status_flags_i <= "1001";
						elsif(cnt = 172) then
							node_status_flags_i <= "0000";
						end if;
					------------------------------------	 
					when 4      =>
                  if(cnt = 12) then --0x0
                     node_status_flags_i <= "0000";
                  elsif(cnt = 23) then --0xf
                     node_status_flags_i <= "1111";
                  elsif(cnt = 50) then --0x5
                     node_status_flags_i <= "1110";
                  elsif(cnt = 61) then --0xa
                     node_status_flags_i <= "0100";
                  elsif(cnt = 72) then --0x9
                     node_status_flags_i <= "0000";
                  elsif(cnt = 86) then --0xd
                     node_status_flags_i <= "0111";
                  elsif(cnt = 99) then --0x6
                     node_status_flags_i <= "0010";
                  end if; 						
					when others =>
						null;
				end case;

				cnt := cnt + 1;
			end if;
		end if;
	end process node_status_flags_interface;	
	
	status_lane_in_interface : process(bp_clk_i)
		variable node_cnt : integer := 0;
		variable node_fifo_status_flags : fifo_status_flags_type(0 to NODES - 1) := (others => (others => '0'));
		variable cnt : integer := 0;
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				node_fifo_status_flags	:= (others => (others => '0'));
				cnt  := 0;
			else
				case (testcase) is
					when 4      =>
						if(cnt = 12) then -- 0x0
							node_fifo_status_flags(0) := "0011";
							node_fifo_status_flags(1) := "0010";
							node_fifo_status_flags(2) := "1001";
							node_fifo_status_flags(4) := "0110";
						elsif(cnt = 23) then -- 0xf
							node_fifo_status_flags(0) := "1001";
							node_fifo_status_flags(1) := "0101";
							node_fifo_status_flags(2) := "1111";
							node_fifo_status_flags(4) := "0101";
						elsif(cnt = 50) then -- 0x5
                     node_fifo_status_flags(0) := "1111";
                     node_fifo_status_flags(1) := "0010";
                     node_fifo_status_flags(2) := "1111";
                     node_fifo_status_flags(4) := "1010";
						elsif(cnt = 61) then -- 0xa
                     node_fifo_status_flags(0) := "0001";
                     node_fifo_status_flags(1) := "0111";
                     node_fifo_status_flags(2) := "0001";
                     node_fifo_status_flags(4) := "1101";
						elsif(cnt = 72) then -- 0x9 
                     node_fifo_status_flags(0) := "1101";
                     node_fifo_status_flags(1) := "1010";
                     node_fifo_status_flags(2) := "0101";
                     node_fifo_status_flags(4) := "0111";
						elsif(cnt = 86) then -- 0xd
                     node_fifo_status_flags(0) := "1111";
                     node_fifo_status_flags(1) := "1001";
                     node_fifo_status_flags(2) := "1111";
                     node_fifo_status_flags(4) := "1101";
						elsif(cnt = 99) then -- 0x6
                     node_fifo_status_flags(0) := "0000";
                     node_fifo_status_flags(1) := "1111";
                     node_fifo_status_flags(2) := "1111";
                     node_fifo_status_flags(4) := "0010";
						end if; 

					when others =>
						null;
				end case;

				status_lanes_in_i(c_status_lanes_width - 1 downto c_status_flags_width) <= std_logic_vector(to_unsigned(node_cnt, 5));
				status_lanes_in_i(c_status_flags_width - 1 downto 0) <= node_fifo_status_flags(node_cnt);
				
				if(node_cnt = 7) then
					node_cnt := 0;
				else 
					node_cnt := node_cnt + 1;
				end if;
				
				cnt := cnt + 1;
				
			end if;
		end if;
	end process status_lane_in_interface;	
	
	status_lane_out_interface : process(bp_clk_i)
		variable node_status_i : std_logic_vector(c_status_flags_width - 1 downto 0) := (others => '0');
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				node_status_i := (others => '0'); 
			else
				case (testcase) is
					when 3      =>
						if(status_lanes_out_i(c_status_lanes_width - 1 downto c_status_flags_width) = std_logic_vector(to_unsigned(CONNECT_ID, 5))) then
							if(node_status_i /= status_lanes_out_i(c_status_flags_width - 1 downto 0)) then
								if(status_lanes_out_i(c_status_flags_width - 1 downto 0) /= node_status_flags_i) then
									report "Status is not being passed onto the status lanes" severity failure;
								end if;
							end if;
							node_status_i	:= status_lanes_out_i(c_status_flags_width - 1 downto 0);
						end if;
						
					when others =>
						null;
				end case;

			end if;
		end if;
	end process status_lane_out_interface;	
	
	dest_fifo_status_interface : process(bp_clk_i)
		variable cnt : integer := 0;
	begin
		if (rising_edge(bp_clk_i)) then
			if (reset_test = '1') then
				cnt := 0;
			else
				case (testcase) is
					when 4 =>
					
						if(cnt = 22 and dest_fifo_status_i /= (not X"0")) then -- 0x0
							report "Dest_fifo_status signals are not working well" severity failure;
						end if;
						
						if(cnt = 33 and dest_fifo_status_i /= (not X"F")) then -- 0xf
                     report "Dest_fifo_status signals are not working well" severity failure;
                  end if;				
                  
						if(cnt = 60 and dest_fifo_status_i /= (not X"5")) then -- 0x5
                     report "Dest_fifo_status signals are not working well" severity failure;
                  end if;				
                  
						if(cnt = 71 and dest_fifo_status_i /= (not X"a")) then -- 0xa
                     report "Dest_fifo_status signals are not working well" severity failure;
                  end if;				
                  
						if(cnt = 82 and dest_fifo_status_i /= (not X"9")) then -- 0x9
                     report "Dest_fifo_status signals are not working well" severity failure;
                  end if;				
                  
						if(cnt = 96 and dest_fifo_status_i /= (not X"D")) then -- 0xd
                     report "Dest_fifo_status signals are not working well" severity failure;
                  end if;				
                  
						if(cnt = 109 and dest_fifo_status_i /= (not X"6")) then -- 0x6
                     report "Dest_fifo_status signals are not working well" severity failure;
                  end if;					
						
					when others =>
						null;
				end case;
				
				cnt := cnt + 1;
				
			end if;
		end if;
	end process dest_fifo_status_interface;	
	

	-- process to check counters before resetting the test
	process(bp_clk_i)
		variable prev_reset_test   : std_logic := '0';
		variable sum_sent_data     : integer   := 0;
		variable sum_received_data : integer   := 0;
	begin
		if (rising_edge(bp_clk_i)) then
			-- at the moment that reset asserts, we want to check the counters before they are resetted.
			if (reset_test = '1' and prev_reset_test = '0') then
				
				-- Calculate the amount of data sent on the lanes
				sum_sent_data	:= 0;
				for i in 0 to CONNECT_LANES - 1 loop
					for j in 0 to NODES - 1 loop
						for k in 0 to 3 loop
							sum_sent_data := sum_sent_data + to_integer(send_to_lanes_i(i)(j)(k));
						end loop;
					end loop; 
				end loop;
				-- Accumulate the f2b interface counters
				for i in 0 to 3 loop
					sum_sent_data := sum_sent_data + to_integer(send_to_f2b(i));
				end loop;
				
				-- Calculate the amount of data received on the lanes
				sum_received_data	:= 0;
				for j in 0 to NODES - 1 loop
					for k in 0 to 3 loop
						sum_received_data := sum_received_data + to_integer(captured_on_lanes_i(j)(k));
					end loop;
				end loop; 
				-- Accumulate the b2f interface counters
				for i in 0 to 3 loop
					sum_received_data := sum_received_data + to_integer(captured_on_b2f(i));
				end loop;				
				
				if(sum_sent_data /= sum_received_data) then
					report "Somewhere, data got lost..." severity failure;
				end if;
				
			end if;
			prev_reset_test := reset_test;
		end if;
	end process;

	lanes_in_sim_i : lane_input_simulator
		generic map(
			CONNECT_LANES => CONNECT_LANES,
			CONNECT_ID    => CONNECT_ID
		)
		port map(
			clk               => bp_clk_i,
			rst               => reset_test,
			lane_mode         => lanes_mode_i,
			lane_destinations => lanes_destination_i,
			lanes             => data_lanes_in_i,
			send_to_lanes     => send_to_lanes_i
		);

	lanes_out_check_i : lane_output_checker
		generic map(
			CONNECT_LANES => CONNECT_LANES,
			CONNECT_ID    => CONNECT_ID
		)
		port map(
			clk               => bp_clk_i,
			rst               => reset_test,
			lane_destinations => lanes_destination_i,
			lanes             => data_lanes_out_i,
			captured_on_lanes => captured_on_lanes_i
		);

	dut : dyplo_connect_slice
		generic map(
			CONNECT_ID    => CONNECT_ID,
			CONNECT_LANES => CONNECT_LANES,
			NODES         => NODES,
			MONITOR       => MONITOR
		)
		port map(
			-- destinations of streams
			stream_destinations => stream_destinations_i,
			-- connection to node
			f2b_tdata           => f2b_tdata_i,
			f2b_tstrb 			  => f2b_tstrb_i,
			f2b_tuser           => f2b_tuser_i,
			f2b_tstream_id      => f2b_tstream_id_i,
			f2b_tvalid          => f2b_tvalid_i,
			f2b_tready          => f2b_tready_i,
			b2f_tdata           => b2f_tdata_i,
			b2f_tstrb 				=> b2f_tstrb,
			b2f_tuser           => b2f_tuser_i,
			b2f_tstream_id      => b2f_tstream_id_i,
			b2f_tvalid          => b2f_tvalid_i,
			-- Backplane
			bp_clk              => bp_clk_i,
			bp_ena              => bp_ena_i,
			status_flags        => node_status_flags_i,
			dest_fifo_status    => dest_fifo_status_i,
			status_lanes_in     => status_lanes_in_i,
			status_lanes_out    => status_lanes_out_i,
			status_error        => status_error_i,
			data_lanes_in       => data_lanes_in_i,
			data_lanes_out      => data_lanes_out_i,
			f2b_count           => f2b_count_i,
			b2f_count           => b2f_count_i,
			bp_count            => bp_count_i
		);

end testbench;
